﻿# Biblos 5 - The Accessible Braille Word Processor
# Components for NVDA 2024.1 (and later?)
# Copyright (C) 2013-2024 Giuseppe Di Grande
# giuseppe@digrande.it - https://www.digrande.it/
# Subscribe https://www.facebook.com/groups/BiblosBraille
# For translators: some text must be translated with Biblos Interface Editor
# Nvda settings text must be translated into "biblosLoca.py"

from controlTypes import role
from NVDAObjects.IAccessible import IAccessible
from NVDAObjects.behaviors import EditableTextWithoutAutoSelectDetection
from NVDAObjects import NVDAObjectTextInfo
from displayModel import DisplayModelTextInfo, EditableTextDisplayModelTextInfo
from scriptHandler import isScriptWaiting
from synthDriverHandler import *
from speech.commands import CallbackCommand, EndUtteranceCommand
from utils.displayString import DisplayStringIntEnum
import api
import appModuleHandler
import braille
import colors
import comHelper
import config
import controlTypes
import displayModel
import eventHandler
import globalCommands
import gui
import logHandler
import louisHelper
import louis
import queueHandler
import scriptHandler
import speech
import synthDriverHandler
import textInfos
import tones
import ui
import unicodedata
import wx
import keyboardHandler
import time
import ctypes
import ctypes.wintypes
import winKernel
import winUser
from . import biblosConf


#***
# The lord of the Rings is a brick that builds universes
class biblosLocales:
	StrBlank = "blank"
	StrInTable = "in table"
	StrOutTable = "out table"
	StrTableRow = "row"
	StrTableCol = "col"
	StrInList = "in list"
	StrOutList = "out list"
	StrListLevel = "level"
	StrCapital = "capital"
	StrTextTransliteration = "Text transliteration"
	StrCharacter = "Character"
	StrNVDASettings = "Biblos Settings for NVDA"
	StrNoSelection = "No selection"
	StrCutToClipboard = "cut"
	StrCopyToClipboard = "copied"
	StrPasteFromClipboard = "pasted"

	bUti = None

	def loadLocales(self,comUti):
		self.bUti = comUti
		self.StrBlank = self.getLocaleStr(0,self.StrBlank)
		self.StrInTable = self.getLocaleStr(1,self.StrInTable)
		self.StrOutTable = self.getLocaleStr(2,self.StrOutTable)
		self.StrTableRow = self.getLocaleStr(3,self.StrTableRow)+" %d"
		self.StrTableCol = self.getLocaleStr(4,self.StrTableCol)+" %d"
		self.StrInList = self.getLocaleStr(5,self.StrInList)
		self.StrOutList = self.getLocaleStr(6,self.StrOutList)
		self.StrListLevel = self.getLocaleStr(7,self.StrListLevel)+" %d"
		self.StrCapital = self.getLocaleStr(8,self.StrCapital)
		self.StrTextTransliteration = self.getLocaleStr(9,self.StrTextTransliteration)
		self.StrCharacter = self.getLocaleStr(10,self.StrCharacter)+" %d"
		self.StrNVDASettings = self.getLocaleStr(11,self.StrNVDASettings)
		self.StrNoSelection = self.getLocaleStr(12,self.StrNoSelection)
		self.StrCutToClipboard = self.getLocaleStr(13,self.StrCutToClipboard)
		self.StrCopyToClipboard = self.getLocaleStr(14,self.StrCopyToClipboard)
		self.StrPasteFromClipboard = self.getLocaleStr(15,self.StrPasteFromClipboard)

	def getLocaleStr(self, ID, default):
		try:
			str = self.bUti.getTextSR(ID)
			if (str is None) or (str==""): return default
			else: return str
		except:
			logHandler.log.info("Utility GetTextSR error.")
			return default


#***
# Stephen King is the best developer of the world
class biblosHandler:

	conf = biblosConf.biblosConfig()
	textTransliteration = False
	sayIndentation = False
	sayBrailleNav = False
	brailleMode = 0
	brailleTole = 7
	oldBraillePos = 0
	oldStartPos = 0
	oldEndPos = 0
	locales = biblosLocales()
	biblosClassNames = ["TBiblosDocument","TRVTableInplaceEdit"]
	biblosFullItems = [-1, -60, -59, -1025, -1029, -1030]
	_bDoc = None
	_bUti = None
	typedWord = []
	curList = 0
	curLevel = 0
	lastText = ""
	oldTable = False
	oldTableCol = 0
	oldTableRow = 0
	newTable = False
	newTableCol = 0
	newTableRow = 0
	tableRow = 0
	tableCol = 0

	@property
	def bDoc(self):
		try:
			if self._bDoc==None:
				self._bDoc = comHelper.getActiveObject("BiblosObj.Document",dynamic=True)
		except: self._bDoc = None
		return self._bDoc

	@property
	def bUti(self):
		try:
			if self._bUti==None:
				self._bUti = comHelper.getActiveObject("BiblosObj.Utility",dynamic=True)
		except: self._bUti = None
		return self._bUti

	def loadConfig(self):
		try:
			self.typedWord = []
			self.locales.loadLocales(self.bUti)
		except:
			logHandler.log.info("Biblos Config locale error.")
		try:
			self.textTransliteration = bHan.conf.textTransliteration
			self.sayIndentation = bHan.conf.sayIndentation
			self.sayBrailleNav = bHan.conf.sayBrailleNav
			self.brailleMode = bHan.conf.brailleMode
			self.brailleTole = bHan.conf.brailleTolerance
		except:
			self.textTransliteration = False
			self.sayIndentation = False
			self.brailleMode = 0
			self.brailleTole = 7

	def isFocused(self):
		obj = api.getCaretObject()
		return obj.windowClassName in self.biblosClassNames

	def isCalculator(self):
		obj = api.getForegroundObject()
		return (obj.windowClassName=="TBiblosCalculator")

	def isTextItem(self):
		item = self.bDoc.GetCurrentItemStyle()
		return item >= 0

	def isFullItem(self):
		item = self.bDoc.GetCurrentItemStyle()
		return item in self.biblosFullItems

	def loadTable(self):
		if self.isFocused():
			self.newTable = self.bDoc.InTable()
			self.newTableCol = self.bDoc.getCurrentTableCol()
			self.newTableRow = self.bDoc.getCurrentTableRow()
		else:
			self.newTable = False
			self.newTableCol = 0
			self.newTableRow = 0

	def saveTable(self):
		self.oldTable = self.newTable
		self.oldTableCol = self.newTableCol
		self.oldTableRow = self.newTableRow

	def biblosTransliterate(self,text):
		try:
			if self.textTransliteration:
				s = self.uniToWide(text)
				s = self.bUti.getTransliterate(s)
				return self.wideToUni(s)
			else:
				return text
		except:
			return text

	def isUniSurr(self,ct,ind):
		try:
			c1 = ord(ct[ind])
			c2 = ord(ct[ind+1])
			return (c1>=0xD800) and (c1<=0xDBFF) and (c2>=0xDC00) and (c2<=0xDFFF)
		except:
			return False

	def uniChr(self,wide,ind):
		h = (ord(wide[ind])-0xD800)<<10
		l = ord(wide[ind+1])-0xDC00
		return chr(h+l)

	def wideToUni(self,wide):
		if (wide is None) or (wide==""): return ""
		rs = ""
		nd = 0
		while nd<len(wide):
			if self.isUniSurr(wide,nd):
				rs+= self.uniChr(wide,nd)
				nd+= 2
			else:
				rs+= wide[nd]
				nd+= 1
		return rs

	def wideChr(self,u):
		c = ord(u)
		if c>=0x10000:
			c = c-0x10000
			h = ((c>>10) & 0x03FF)+ 0xD800
			l = (c & 0x03FF)+ 0xDC00
			return chr(h)+chr(l)
		else:
			return chr(c)

	def uniToWide(self,uni):
		if (uni is None) or (uni==""): return ""
		rs = ""
		for c in uni:
			rs+= self.wideChr(c)
		return rs

# Speak a char
# info is textInfo object. Low priority
# text is string with a char. High priority
	def biblosSpeakCharacter(self,info,text):
		if text is not None:
			ct = text
		elif (info is not None):
			ct = info.text
		else:
			ct = ""
		if (ct is None) or (len(ct)==0) or ((len(ct)==1) and (ord(ct)<32)):
			speech.speakText(self.locales.StrBlank)
			return
		if len(ct)>=2:
			speech.speakText(self.biblosTransliterate(ct))
			return
		cn = self.bUti.GetCharName(self.wideChr(ct))
		if cn:
			synth=synthDriverHandler.getSynth()
			synthConfig = config.conf["speech"][synth.name]
			u = ct.isupper()
			if u and synth.isSupported("pitch") and synthConfig["capPitchChange"]:
				oldPitch = synthConfig["pitch"]
				synth.pitch = max(0, min(oldPitch+synthConfig["capPitchChange"], 100))
			speech.speakText(cn)
			if u and synth.isSupported("pitch") and synthConfig["capPitchChange"]:
				synth.pitch=oldPitch
			if u and synthConfig["sayCapForCapitals"]:
				speech.speakText(self.locales.StrCapital)
			if u and  synthConfig["beepForCapitals"]:
				tones.beep(2000,50)
		else:
			speech.speakText(self.locales.StrCharacter % ord(ct[0]))

	def biblosSpeakCurrentCharacter(self,level):
		obj = api.getCaretObject()
		try:
			info = obj.makeTextInfo(textInfos.POSITION_CARET)
		except:
			self.biblosSpeakObject(obj)
			return
		if self.isFocused()==False:
			info.expand(textInfos.UNIT_CHARACTER)
			self.biblosSpeakInfos(False)
		else:
			self.biblosSpeakInfos(True)
			if self.isTextItem()==False:
				self.biblosSpeakItem(level)
				return
			info.text = self.wideToUni(self.bDoc.GetCurrentChar())
		if level==0:
			self.biblosSpeakCharacter(info,None)
		elif level==1:
				speech.speakSpelling(info.text,useCharacterDescriptions=True)
		else:
			try:
				c = ord(info.text)
				speech.speakText("%d," % c)
				speech.speakSpelling(hex(c))
			except:
				speech.speakTextInfo(info,unit=textInfos.UNIT_CHARACTER)

	def biblosSpeakCharacterEx(self):
		if self.isFocused()==False:
			obj = api.getCaretObject()
			if obj.role==controlTypes.Role.MENUITEM: return
			try:
				info = obj.makeTextInfo(textInfos.POSITION_CARET)
			except:
				self.biblosSpeakObject(obj)
				return
			info.expand(textInfos.UNIT_CHARACTER)
			self.biblosSpeakInfos(False)
			self.biblosSpeakCharacter(info,None)
			return
#
		self.biblosSpeakInfos(True)
		if self.bDoc.IsEndOfItem():
			self.biblosSpeakCharacter(None,"")
			return
		if self.isTextItem()==False:
			self.biblosSpeakItem(0)
			return
		s = self.wideToUni(self.bDoc.GetCurrentChar())
		if s is None: s = ""
		self.biblosSpeakCharacter(None,s)

	def biblosSpeakItem(self,deep):
		info = self.bDoc.GetCurrentItemInfo(deep)
		speech.speakText(info)
 
	def getStringIndentation(self, s):
		i = 0
		while i<len(s):
			if ord(s[i])>32: return str(i)
			i+= 1
		return ""

	def biblosSpeakLine(self,text):
		if (text is None) or (len(text)==0):
			speech.speakText(self.locales.StrBlank)
			return
		if self.sayIndentation:
			speech.speakText(self.getStringIndentation(text))
		if len(text) == 1:
			self.biblosSpeakCharacter(None,text)
			return
		cn = self.biblosTransliterate(text)
		speech.speakText(cn)

	def biblosSpeakLineEx(self):
		obj = api.getCaretObject()
		if self.isFocused():
			s = self.wideToUni(self.bDoc.GetCurrentLine())
			if (s is None): s = ""
			self.biblosSpeakLine(s)
		else:
			if obj.role==controlTypes.Role.MENUITEM: return
			try:
				info = obj.makeTextInfo(textInfos.POSITION_CARET)
				info.expand(textInfos.UNIT_LINE)
				self.biblosSpeakInfos(False)
				self.biblosSpeakLine(info.text)
			except:
				pass

	def biblosSpeakParagraph(self,text):
		if (text is None) or (len(text)==0):
			speech.speakText(self.locales.StrBlank)
			return
		if len(text) == 1:
			self.biblosSpeakCharacter(None,text)
			return
		cn = self.biblosTransliterate(text)
		speech.speakText(cn)

	def biblosSpeakParagraphEx(self):
		obj = api.getCaretObject()
		if self.isFocused():
			self.biblosSpeakInfos(True)
			s = self.wideToUni(self.bDoc.GetCurrentPara())
			if s is None: s = ""
			self.biblosSpeakParagraph(s)
		else:
			if obj.role==controlTypes.Role.MENUITEM: return
			info = obj.makeTextInfo(textInfos.POSITION_CARET)
			info.expand(textInfos.UNIT_PARAGRAPH)
			self.biblosSpeakInfos(False)
			self.biblosSpeakParagraph(info.text)

	def biblosSpeakWord(self,text):
		if (text is None) or (len(text)==0):
			speech.speakText(self.locales.StrBlank)
			return
		if len(text) == 1:
			self.biblosSpeakCharacter(None,text)
			return
		cn = self.biblosTransliterate(text)
		speech.speakText(cn)

	def biblosSpeakWordEx(self):
		if self.isFocused()==False:
			obj = api.getCaretObject()
			if obj.role==controlTypes.Role.MENUITEM:
				return
			info = obj.makeTextInfo(textInfos.POSITION_CARET)
			info.expand(textInfos.UNIT_WORD)
			self.biblosSpeakInfos(False)
			self.biblosSpeakWord(info.text)
			return
		self.biblosSpeakInfos(True)
		if self.isTextItem()==False:
			self.biblosSpeakItem(0)
			return
		s = self.wideToUni(self.bDoc.GetCurrentWord())
		if s is None: s = "" 
		self.biblosSpeakWord(s)

	def biblosSpeakObject(self,obj):
		self.biblosSpeakInfos(False)
		speech.speakObjectProperties(obj)
		try: n = obj.name
		except: n = "--"
		if (n is not None) and (len(n)==1):
			self.biblosSpeakCharacter(None,n)
			return
		cn = self.biblosTransliterate(n)
		speech.speakText(cn)

	def biblosSpeakSelection(self,text):
		stext = ""
		if len(text) > 1:
			stext = self.biblosTransliterate(text)
		else:
			cn = self.bUti.GetCharName(text)
			if cn:
				stext = cn
			else:
				stext = text
		speech.speakSelectionMessage(_("selected %s"), stext)

	def biblosSpeakFontInfo(self):
		if self.isFocused():
			tc = self.bDoc.GetCurrentTextColor( )
			bc = self.bDoc.GetCurrentBackColor( )
			s = self.bDoc.GetCurrentFontInfo(1)
			s = s.replace("%tc", colors.RGB(tc&0xff,(tc>>8)&0xff,(tc>>16)&0xff).name)
			s = s.replace("%bc", colors.RGB(bc&0xff,(bc>>8)&0xff,(bc>>16)&0xff).name)
			speech.speakText(s)
			braille.handler.message(s)
		else:
			globalCommands.commands.script_reportFormatting(None)

	def biblosSpeakParaInfo(self):
		if self.isFocused():
			s = self.bDoc.GetCurrentParaInfo(1)
			speech.speakText(s)
			braille.handler.message(s)

	def biblosSpeakBrailleInfo(self):
		if self.isFocused():
			s = self.bDoc.GetCurrentBrailleInfo(1)
			speech.speakText(s)
			braille.handler.message(s)

	def biblosSpeakDocumentInfo(self,level):
		if self.isFocused():
			s = self.bDoc.GetCurrentDocumentInfo(level)
			speech.speakText(s)
			braille.handler.message(s)

	def biblosSpeakInfos(self,focused):
		if focused == False:
			self.curList = 0
			self.curLevel = 0
			self.lastText = ""
			return
		self.lastText = self.wideToUni(self.bDoc.GetCurrentItemText())
		cl = self.bDoc.GetCurrentList()
		cm = self.bDoc.GetCurrentListLevel()
		if (self.curList == cl):
			if (cl == 0):
				return
			if (self.curLevel == cm):
				return
			speech.speakText(self.locales.StrListLevel % cm)
			self.curLevel = cm
			return
		else:
			if (cl == 0):
				speech.speakText(self.locales.StrOutList)
				self.curList = 0
				self.curLevel = 0
				return
			speech.speakText(self.locales.StrInList)
			speech.speakText(self.locales.StrListLevel % cm)
			self.curList = cl
			self.curLevel = cm

	def biblosBrailleSettings(self):
		if self.isFocused()==False: return
		opt = 0
		if self.conf.brailleFormatting: opt|= 1
		if self.conf.brailleTableCoords: opt|= 2
		if self.conf.brailleExtras: opt|= 4
		try: self.bDoc.SetBrailleOptions(opt)
		except: pass

	def biblosBrailleLine(self):
		if braille.handler.enabled:
			if self.isFocused()==False: return
			obj = api.getCaretObject()
			region = BiblosRegion(obj)
			region.updateLine()
			region.updateCursor()
			self.biblosBrailleRefresh(region)

	def biblosBraillePara(self):
		if braille.handler.enabled:
			if self.isFocused()==False: return
			obj = api.getCaretObject()
			region = BiblosRegion(obj)
			region.updatePara()
			region.updateCursor()
			self.biblosBrailleRefresh(region)

	def biblosBrailleRefresh(self,region):
		if braille.handler._messageCallLater:
			braille.handler._messageCallLater.Stop()
			braille.handler._messageCallLater = None
		self.oldBraillePos = braille.handler.buffer.cursorPos
		if self.oldBraillePos==None: self.oldBraillePos = 0
		self.oldStartPos = braille.handler.buffer.windowStartPos
		if self.oldStartPos==None: self.oldStartPos = 0
		self.oldEndPos = braille.handler.buffer.windowEndPos
		if self.oldEndPos==None: self.oldEndPos = 0
# Now we start dancing
		buffer = BiblosBrailleBuffer(braille.handler)
		braille.handler.buffer = buffer
		buffer.clear()
		buffer.regions.append(region)
		region.update()
		buffer.update()
		braille.handler.update()

	def biblosBrailleRestore(self):
		braille.handler.buffer = braille.handler.mainBuffer
		braille.handler.buffer.update()
		braille.handler.update()

	def biblosTableChange(self):
		if self.isFocused()==False:
			self.newTable = False
			return
#
		self.loadTable()
		if self.oldTable==False and self.newTable==False: return
# Round 1: fight!
		if self.oldTable==True and self.newTable==False:
			speech.speakText(self.locales.StrOutTable)
			self.saveTable()
			return
# Round 2: Fight!
		if self.oldTable==False and self.newTable==True:
			speech.speakText(self.locales.StrInTable)
			self.oldTableCol = 0
			self.oldTableRow = 0
# Wonder boy is alive
		if self.oldTableCol==self.newTableCol and self.oldTableRow==self.newTableRow:
			return
# The boss
		if self.oldTableCol!=self.newTableCol:
			speech.speakText(self.locales.StrTableCol % self.newTableCol)
		if self.oldTableRow!=self.newTableRow:
			speech.speakText(self.locales.StrTableRow % self.newTableRow)
# Thank you Mr Tramiel
		self.saveTable()

	def biblosTabKey(self):
		if self.isFocused()==False:
			return
		if self.bDoc.ReadOnly()==True:
			return
		self.biblosSpeakInfos(True)
		if (self.bDoc.InList()==True) and (self.bDoc.IsStartOfLine()==True):
			self.biblosSpeakLineEx()
			return
		if self.bDoc.InTable()==True:
			return
		self.biblosSpeakDocumentInfo(3)


	def getBiblosTypedCharacter(self,ch):
		if self.isFocused()==False:
			return ch
		if self.lastText is None: self.lastText = ""
		lt = self.wideToUni(self.bDoc.GetCurrentItemText(self.uniToWide(self.lastText)))
		self.lastText = self.wideToUni(self.bDoc.GetCurrentItemText())
		if (lt is None): lt = ""
		return lt


	def biblosSpeakTypedWord(self,word):
		if (word is not None) and (len(word)>0):
			speech.speakText(self.biblosTransliterate(word))


	def biblosEnterKey(self):
		if self.isFocused():
			self.biblosSpeakLineEx()


	def biblosBackspaceKey(self):
		if self.isFocused():
			s = self.wideToUni(self.bDoc.GetCurrentItemBack())
			if (s is None) or (len(s)==0): tones.beep(440,50)
			else: self.biblosSpeakLine(s)


	def biblosSpeakAll(self):
		if self.isFocused()==False: return
		parag = self.wideToUni(self.bDoc.GetCurrentParaFromCaret())
		api.processPendingEvents(processEventQueue=False)
		self.biblosSpeakAllText(parag)

	def biblosSpeakAllNext(self):
		if self.bDoc.NextParagraph(1):
			parag = self.wideToUni(self.bDoc.GetCurrentPara())
		else:
			tones.beep(216,200)
			return
		api.processPendingEvents(processEventQueue=False)
		self.biblosSpeakAllText(parag)

	def biblosSpeakAllText(self, parag):
		cb = CallbackCommand(self.biblosSpeakAllNext, name="say-all:next")
		sq = [self.biblosTransliterate(parag),cb]
		speech.speak(sq)
		self.biblosBraillePara()
		if (parag is None) or (len(parag)==0):
			tones.beep(3000,20)
			time.sleep(0.4)


# *****
# Alone in the dark
bHan = None

#***
# Mommy says panties need to be changed every day
class AppModule(appModuleHandler.AppModule):
	foreground = None
	curHandle = 0

	def __init__(self, *args, **kwargs):
		global bHan
		super(AppModule, self).__init__(*args, **kwargs)
		bHan = biblosHandler()
		logHandler.log.info("Biblos Object initialized.")

	def event_gainFocus(self, obj, nextHandler):
		nextHandler()

	def chooseNVDAObjectOverlayClasses(self, obj, clsList):
		if obj.windowClassName == "TBiblosDocument":
			clsList.insert(0, BiblosDocument)
		elif obj.windowClassName == "TRVTableInplaceEdit":
			clsList.insert(0, BiblosDocument)

	def script_sayCharacter(self,gesture):
		global bHan
		gesture.send()
		if isScriptWaiting() or eventHandler.isPendingEvents("gainFocus"): return
		api.processPendingEvents(processEventQueue=False)
		bHan.typedWord = []
		hwnd = bHan.bDoc.GetCurrentBiblosId()
		if hwnd==self.curHandle:
			bHan.biblosSpeakCharacterEx()
			bHan.biblosBrailleLine()

	def script_sayCurrentCharacter(self,gesture):
		global bHan
		bHan.typedWord = []
		bHan.biblosSpeakCurrentCharacter(scriptHandler.getLastScriptRepeatCount())

	def script_sayWord(self,gesture):
		global bHan
		gesture.send()
		if isScriptWaiting() or eventHandler.isPendingEvents("gainFocus"): return
		api.processPendingEvents(processEventQueue=False)
		bHan.typedWord = []
		hwnd = bHan.bDoc.GetCurrentBiblosId()
		if hwnd==self.curHandle:
			bHan.biblosSpeakWordEx()
			bHan.biblosBrailleLine()

	def script_sayCurrentWord(self,gesture):
		global bHan
		bHan.typedWord = []
		bHan.biblosSpeakWordEx()

	def script_sayLine(self,gesture):
		global bHan
		oldH = self.curHandle
		gesture.send()
		api.processPendingEvents(processEventQueue=True)
		bHan.typedWord = []
		newH = self.curHandle
		if oldH==newH:
			bHan.biblosSpeakLineEx()
			bHan.biblosBrailleLine()

	def script_sayCurrentLine(self,gesture):
		global bHan
		bHan.typedWord = []
		bHan.biblosSpeakLineEx()

	def script_sayParagraph(self,gesture):
		global bHan
		gesture.send()
		if isScriptWaiting() or eventHandler.isPendingEvents("gainFocus"): return
		api.processPendingEvents(processEventQueue=False)
		bHan.typedWord = []
		bHan.biblosSpeakParagraphEx()
		bHan.biblosBraillePara()

	__gestures = {
		"kb:leftarrow": "sayCharacter",
		"kb:rightarrow": "sayCharacter",
		"kb:home": "sayCharacter",
		"kb:end": "sayCharacter",
		"kb:delete": "sayCharacter",
		"kb:numpad2": "sayCurrentCharacter",
		"kb:uparrow": "sayLine",
		"kb:downarrow": "sayLine",
		"kb:numpad8": "sayCurrentLine",
		"kb:control+leftarrow": "sayWord",
		"kb:control+rightarrow": "sayWord",
		"kb:numpad5": "sayCurrentWord",
		"kb:control+uparrow": "sayParagraph",
		"kb:control+downarrow": "sayParagraph",
		"kb:control+home": "sayLine",
		"kb:control+end": "sayLine",
		"kb:pageUp": "sayLine",
		"kb:pageDown": "sayLine",
		"kb:NVDA+uparrow": "sayCurrentLine",
	}


#***
# This class is used by elephants when they march in herds
class BiblosBrailleItem:
	charPos: int  # Char position
	itemNo: int
	offsetNo: int
	typeform: int  # Liblouis typeform info
	rawText: str  # text char
	cells: list  # Translated braille

	def __init__(self):
		self.charPos = 0
		self.itemNo = 0
		self.offsetNo = 0
		self.typeform = 0
		self.rawText = ""
		self.cells = []

#***
# This class likes to grow lemons for vegans
class BiblosRegion(braille.Region):

	items: list

	def __init__(self, obj):
		self.obj = obj
		self.clear()

	def clear(self):
		self.items = []
		self.rawText = ""
		self.rawTextTypeforms = None
		self.cursorPos = None
		self.selectionStart = None
		self.selectionEnd = None
		self.brailleCells = []
		self.rawToBraillePos = []
		self.brailleToRawPos = []
		self.brailleCursorPos: Optional[int] = None
		self.brailleSelectionStart = None
		self.brailleSelectionEnd = None
		self.hidePreviousRegions = True
		self.focusToHardLeft = False

	def getWChr(self,data,index):
		return chr(int(data[index:index+4],16))

	def getWInt(self,data,index):
		return int(data[index:index+4],16)

	def addBrlItem(self, index, data):
		if index>=len(data): return index
		ni = BiblosBrailleItem()
		l = 0
		try:
			ni.charPos = self.getWInt(data,index)
			index+= 4
			ni.itemNo = self.getWInt(data,index)
			index+= 4
			ni.offsetNo = self.getWInt(data,index)
			index+= 4
			ni.typeform = self.getWInt(data,index)
			index+= 4
			l = self.getWInt(data,index)
			index+= 4
			for x in range(l):
				ni.rawText+= self.getWChr(data,index)
				index+= 4
		except:
			pass
		self.items.append(ni)
		return index

	def itemToRaw(self,item):
		r = 0
		for i in self.items:
			if item==i.charPos: return r
			r+= len(i.rawText)
		return len(self.rawText)

	def rawToItem(self,raw):
		n = 0
		r = 0
		raw+= 1
		while n<len(self.items):
			r+= len(self.items[n].rawText)
			if (r>=raw) and (self.items[n].charPos>0): return n
			n+= 1
		return None

	def styleToTypeform(self,style,formatConfig):
		r = louis.plain_text
		if not formatConfig["reportFontAttributes"]: return r
		if style & 1 == 1: r|= louis.bold
		if style & 2 == 2: r|= louis.italic
		if style & 4 == 4: r|= louis.underline
		return r

	def updateData(self, data):
		self.items = []
		if data==None: data = ""
		i = 0
		while i<len(data):
			i = self.addBrlItem(i,data)
#
		self.rawText = ""
		self.rawTextTypeforms = []
		fc = config.conf["documentFormatting"]
		for n in self.items:
			self.rawText+= bHan.wideToUni(n.rawText)
			self.rawTextTypeforms.append(self.styleToTypeform(n.typeform,fc))

	def updateLine(self):
		global bHan
		try:
			data = bHan.bDoc.GetBrailleCurrentLine()
		except:
			data = ""
		self.updateData(data)

	def updatePara(self):
		global bHan
		try:
			data = bHan.bDoc.GetBrailleCurrentPara()
		except:
			data = ""
		self.updateData(data)

	def updateCursor(self):
		global bHan
		try:
			data = bHan.bDoc.GetBrailleCurrentCursor()
		except:
			data = ""
		try:
			self.cursorPos = self.itemToRaw(self.getWInt(data,0))
		except:
			self.cursorPos = None
		try:
			self.selectionStart = self.itemToRaw(self.getWInt(data,4))
			self.selectionEnd = self.itemToRaw(self.getWInt(data,8))
		except:
			self.selectionStart = None
			self.selectionEnd = None

	def update(self):
		super(BiblosRegion,self).update()

	def routeTo(self, braillePos):
		global bHan
		try:
			rp = self.brailleToRawPos[braillePos]
			if rp==None: return
			cp = self.rawToItem(rp)
			if cp==None: return
			bHan.bDoc.SetBrailleCurrentCursor(self.items[cp].itemNo, self.items[cp].offsetNo)
			self.updateCursor()
			if bHan.sayBrailleNav:
				queueHandler.queueFunction(queueHandler.eventQueue, bHan.biblosSpeakCharacterEx)
			queueHandler.queueFunction(queueHandler.eventQueue, bHan.biblosBrailleLine)
		except: pass

	def nextLine(self):
		global bHan
		try: r = bHan.bDoc.NextLine(1)
		except: r = False
		if r:
			if bHan.sayBrailleNav: queueHandler.queueFunction(queueHandler.eventQueue, bHan.biblosSpeakLineEx)
			queueHandler.queueFunction(queueHandler.eventQueue, bHan.biblosBrailleLine)
		return r

	def previousLine(self, start=False):
		global bHan
		try:
			if start: r = bHan.bDoc.PriorLine(1)
			else: r = bHan.bDoc.PriorLine(2)
		except: r = False
		if r:
			if bHan.sayBrailleNav: queueHandler.queueFunction(queueHandler.eventQueue, bHan.biblosSpeakLineEx)
			queueHandler.queueFunction(queueHandler.eventQueue, bHan.biblosBrailleLine)
		return r


#***
# This class has breakfast with a croissant
class BiblosBrailleBuffer(braille.BrailleBuffer):

	_windowEndPos = 0

	def _get_windowEndPos(self):
		return self._windowEndPos

	def _set_windowEndPos(self,endPos):
		self._windowEndPos = endPos

	def update(self):
		global bHan
		super(BiblosBrailleBuffer,self).update()
		if bHan.brailleMode==0: self.updateWindowBlock()
		else: self.updateWindowScroll()

	def _nextWindow(self):
		global bHan
		oldStart = self.windowStartPos
		oldEnd = self.windowEndPos
		bCells = len(self.brailleCells)
		if bHan.brailleMode==0:
			if config.conf["braille"]["wordWrap"]: segs=self.getBrailleWordWrapSegments()
			else: segs=self.getBrailleSimpleSegments()
			cur = self.getBrailleWindowFromPos(segs, oldStart) + 1
			if cur>=len(segs): return False
			ws = self.getBrailleWindowStartFrom(segs, cur)
			we = self.getBrailleWindowEndFrom(segs, cur)
		elif bHan.brailleMode==1:
			ws = self.windowStartPos+self.handler.displaySize
			we = ws+self.handler.displaySize
			if we>bCells: we=bCells
			ws = we- self.handler.displaySize
			if ws<0: ws=0
		self.windowStartPos = ws
		self.windowEndPos = we
		return ws!=oldStart or we!=oldEnd

	def _previousWindow(self):
		global bHan
		oldStart = self.windowStartPos
		oldEnd = self.windowEndPos
		bCells = len(self.brailleCells)
		if bHan.brailleMode==0:
			if config.conf["braille"]["wordWrap"]: segs=self.getBrailleWordWrapSegments()
			else: segs=self.getBrailleSimpleSegments()
			cur = self.getBrailleWindowFromPos(segs, oldStart) -1
			if cur<0: return False
			ws = self.getBrailleWindowStartFrom(segs, cur)
			we = self.getBrailleWindowEndFrom(segs, cur)
		elif bHan.brailleMode==1:
			ws = self.windowStartPos-self.handler.displaySize
			if ws<0: ws=0
			we = ws+self.handler.displaySize
			if we>bCells: we=bCells
		self.windowStartPos = ws
		self.windowEndPos = we
		return ws!=oldStart or we!=oldEnd

	def _nextLine(self):
		if len(self.regions)>0:
			self.regions[0].nextLine()

	def _previousLine(self):
		if len(self.regions)>0:
			self.regions[0].previousLine()

	def scrollForward(self):
		if self._nextWindow(): self.updateDisplay()
		else: self._nextLine()

	def scrollBack(self):
		if self._previousWindow(): self.updateDisplay()
		else: self._previousLine()

	def updateWindowBlock(self):
		bCells = len(self.brailleCells)
		if self.handler.displaySize>=bCells:
			self.windowStartPos = 0
			self.windowEndPos = bCells
			return
# I must confess that up until now I have been a happy man
		if config.conf["braille"]["wordWrap"]: segs=self.getBrailleWordWrapSegments()
		else: segs=self.getBrailleSimpleSegments()
# Now I'm eating some tofu and some vegan cheese
		self.windowStartPos = self.getBrailleCurrentWindowStart(segs)
		self.windowEndPos = self.getBrailleCurrentWindowEnd(segs)

	def updateWindowScroll(self):
		global bHan
		bCells = len(self.brailleCells)
		if self.cursorPos==bHan.oldBraillePos:
			ws = bHan.oldStartPos
			we = bHan.oldEndPos
		elif self.cursorPos>bHan.oldBraillePos:
			we = bHan.oldEndPos
			if we-self.cursorPos<bHan.brailleTole: we = self.cursorPos+bHan.brailleTole
			if we>bCells: we=bCells
			ws = we-self.handler.displaySize
			if ws<0: ws=0
		elif self.cursorPos<bHan.oldBraillePos:
			ws = bHan.oldStartPos
			if self.cursorPos-ws<bHan.brailleTole: ws = self.cursorPos-bHan.brailleTole
			if ws<0: ws=0
			we = ws+self.handler.displaySize
			if we>bCells: we=bCells
# Ghosts'n Goblins
		if we-ws<self.handler.displaySize and we<bCells:
			we = ws+self.handler.displaySize
			if we>bCells: we=bCells
		self.windowStartPos = ws
		self.windowEndPos = we

	def getBrailleSomething(self, startPos, endPos):
		if len(self.brailleCells)==0: return None
		if startPos<0: startPos=0
		elif startPos>=len(self.brailleCells): startPos=len(self.brailleCells)-1
		if endPos<0: endPos=0
		elif endPos>=len(self.brailleCells): endPos=len(self.brailleCells)-1
		if startPos<endPos: i= 1
		elif startPos>endPos: i= -1
		else: return startPos
		for index in range(startPos,endPos,i):
			if self.brailleCells[index]!=0: return index
		return startPos

	def getBrailleSpace(self, startPos, endPos):
		if len(self.brailleCells)==0: return None
		if startPos<0: startPos=0
		elif startPos>=len(self.brailleCells): startPos=len(self.brailleCells)-1
		if endPos<0: endPos=0
		elif endPos>=len(self.brailleCells): endPos=len(self.brailleCells)-1
		if startPos<endPos: i= 1
		elif startPos>endPos: i= -1
		else: return startPos
		for index in range(startPos,endPos,i):
			if self.brailleCells[index]==0: return index
		return startPos

	def getBrailleSimpleSegments(self):
		segs = []
		bCells = len(self.brailleCells)
		dSize = self.handler.displaySize
		bIndex = 0
		while True:
			segs.append(bIndex)
			bIndex+= dSize
			if bIndex>=bCells: break
		return segs

	def getBrailleWordWrapSegments(self):
		global bHan
		segs = []
		bCells = len(self.brailleCells)
		dSize = self.handler.displaySize
		bIndex = 0
		while True:
			segs.append(bIndex)
			lIndex = bIndex
			bIndex+= dSize
			if bIndex>=bCells: break
# My first is the west and David Crane is the best
			if self.brailleCells[bIndex]==0: continue
			nIndex = self.getBrailleIndexFromSomething(bIndex, lIndex)
			if abs(bIndex-nIndex)<=bHan.brailleTole: bIndex = nIndex
# Turrican is the hero and Pitfall is the god
		return segs

	def getBrailleIndexFromSomething(self, index, min):
		for n in range(index-1, min, -1):
			if self.brailleCells[n]==0: return n+1
		return index

	def getBrailleWindowFromPos(self, segs, pos):
		l = len(segs)
		if l<=1: return 0
		while l>0:
			l-= 1
			if pos>=segs[l]: return l
		return 0

	def getBrailleWindowStartFrom(self, segs, cur):
		l = len(segs)
		if cur<0: return 0
		if cur>=l: return self.getBrailleWindowStartFrom(segs, l-1)
		return segs[cur]

	def getBrailleWindowEndFrom(self, segs, cur):
		l = len(segs)
		if cur<0: return self.getBrailleWindowEndFrom(segs, 0)
		if cur>=l-1: return len(self.brailleCells)
		return segs[cur+1]

	def getBrailleCurrentWindowStart(self, segs):
		w = self.getBrailleWindowFromPos(segs, self.cursorPos)
		return self.getBrailleWindowStartFrom(segs, w)

	def getBrailleCurrentWindowEnd(self, segs):
		w = self.getBrailleWindowFromPos(segs, self.cursorPos)
		return self.getBrailleWindowEndFrom(segs, w)


#***
# This class explores sidereal spaces like the Star Trek spaceship
class BiblosDocument(EditableTextWithoutAutoSelectDetection):

	def _get_role(self):
		return controlTypes.Role.EDITABLETEXT

	def _getRoleText(self):
		return role._roleLabels[self.role]

	def _get_states(self):
		states=super(BiblosDocument,self).states
		states|= {controlTypes.State.MULTILINE}
		return states

#	def makeTextInfo(self,position):
#		try:
#			t = super(BiblosDocument,self).makeTextInfo(position)
#			t.text = " "
#			return t
#		except: return super(BiblosDocument,self).makeTextInfo(position)

	def event_gainFocus(self):
		global bHan
		api.processPendingEvents(processEventQueue=True)
		bHan.loadConfig()
		bHan.loadTable()
		bHan.biblosBrailleSettings()
		self.invalidateCache()
		bHan.typedWord = []
		fore = api.getForegroundObject()
		bHan.biblosTableChange()
		bHan.biblosBrailleLine()
		newH = bHan.bDoc.GetCurrentBiblosId()
		if newH==self.appModule.curHandle: return
		self.appModule.curHandle = newH
		speech.speakText(self._getRoleText())
		bHan.biblosSpeakInfos(False)
		bHan.biblosSpeakInfos(True)
		bHan.biblosSpeakLineEx()

	def event_loseFocus(self):
		global bHan
		self.appModule.curHandle = 0
		bHan.biblosBrailleRestore()
		super().event_loseFocus()

	def event_typedCharacter(self,ch):
		global bHan
		bHan.biblosBrailleLine()
		rc = bHan.getBiblosTypedCharacter(ch)
		if unicodedata.category(ch)[0] in "LMN":
			bHan.typedWord.append(rc)
		elif ch == "\b":
			del bHan.typedWord[-1:]
		elif ch == u'\u007f':
			return
		elif len(bHan.typedWord)>0:
			w = "".join(bHan.typedWord)
			bHan.typedWord = []
			if config.conf["keyboard"]["speakTypedWords"]:
				bHan.biblosSpeakTypedWord(w)
		if config.conf["keyboard"]["speakTypedCharacters"] and (ord(ch)>=32) and (rc is not None) and (len(rc)>0):
			bHan.biblosSpeakCharacter(None,rc)

	def event_caret(self):
		pass

	def script_cutToClipboard(self,gesture):
		global bHan
		bHan.typedWord = []
		if bHan.bDoc.SelectionExists()==False:
			ui.message(bHan.locales.StrNoSelection)
			return
		gesture.send()
		if isScriptWaiting() or eventHandler.isPendingEvents("gainFocus"): return
		api.processPendingEvents(processEventQueue=False)
		ui.message(bHan.locales.StrCutToClipboard)
		bHan.biblosBrailleLine()

	def script_copyToClipboard(self,gesture):
		global bHan
		bHan.typedWord = []
		if bHan.bDoc.SelectionExists()==False:
			ui.message(bHan.locales.StrNoSelection)
			return
		gesture.send()
		if isScriptWaiting() or eventHandler.isPendingEvents("gainFocus"): return
		api.processPendingEvents(processEventQueue=False)
		ui.message(bHan.locales.StrCopyToClipboard)

	def script_pasteFromClipboard(self,gesture):
		global bHan
		bHan.typedWord = []
		i = bHan.bDoc.GetCurrentItemNo()
		t = bHan.bDoc.GetCurrentItemText()
		gesture.send()
		if isScriptWaiting() or eventHandler.isPendingEvents("gainFocus"): return
		api.processPendingEvents(processEventQueue=False)
		if (i!=bHan.bDoc.GetCurrentItemNo()) or (t!=bHan.bDoc.GetCurrentItemText()):
			ui.message(bHan.locales.StrPasteFromClipboard)
			bHan.biblosBrailleLine()

	def script_sayBackspaceKey(self,gesture):
		global bHan
		bHan.biblosBackspaceKey()
		gesture.send()
		if isScriptWaiting() or eventHandler.isPendingEvents("gainFocus"): return
		api.processPendingEvents(processEventQueue=False)
		bHan.biblosBrailleLine()

	def script_sayF1Key(self,gesture):
		global bHan
		bHan.typedWord = []
		gesture.send()
		if isScriptWaiting() or eventHandler.isPendingEvents("gainFocus"): return
		api.processPendingEvents(processEventQueue=False)
		if bHan.bDoc.SelectionExists():
			bHan.biblosSpeakInfos(False)
			sel = bHan.wideToUni(bHan.bDoc.GetCurrentSel())
			bHan.biblosSpeakLine(sel)
		bHan.biblosBrailleLine()

	def script_sayAll(self,gesture):
		global bHan
		bHan.typedWord = []
		if isScriptWaiting() or eventHandler.isPendingEvents("gainFocus"): return
		api.processPendingEvents(processEventQueue=False)
		bHan.biblosSpeakAll()

	def script_sayLine(self,gesture):
		self.appModule.script_sayLine(gesture)

	def script_sayWord(self,gesture):
		self.appModule.script_sayWord(gesture)

	def script_sayChangeSelection(self,gesture):
		global bHan
		bHan.typedWord = []
		oldText = bHan.bDoc.GetCurrentSel()
		oldIndex = bHan.bDoc.GetCurrentItemNo()
		if oldText == None: oldText = ""
		gesture.send()
		if isScriptWaiting() or eventHandler.isPendingEvents("gainFocus"): return
		api.processPendingEvents(processEventQueue=False)
		newText = bHan.bDoc.GetCurrentSel(oldText)
		newIndex = bHan.bDoc.GetCurrentItemNo()
		if (oldText!=newText) or (oldIndex!=newIndex):
			bHan.biblosSpeakInfos(False)
			bHan.biblosSpeakSelection(bHan.wideToUni(newText))
			bHan.biblosBrailleLine()

	def script_sayEnterKey(self,gesture):
		global bHan
		gesture.send()
		if isScriptWaiting() or eventHandler.isPendingEvents("gainFocus"): return
		api.processPendingEvents(processEventQueue=False)
		if bHan.isCalculator(): bHan.biblosEnterKey()
		bHan.biblosBrailleLine()

	def script_sayTabKey(self,gesture):
		global bHan
		gesture.send()
		if isScriptWaiting() or eventHandler.isPendingEvents("gainFocus"): return
		api.processPendingEvents(processEventQueue=False)
		bHan.biblosTabKey()
		bHan.biblosBrailleLine()

	def script_sayTextInfo(self,gesture):
		global bHan
		bHan.typedWord = []
		scriptCount = scriptHandler.getLastScriptRepeatCount()
		if scriptCount==0:
			bHan.biblosSpeakFontInfo()
		elif scriptCount==1:
			bHan.biblosSpeakParaInfo()
		elif scriptCount==2:
			bHan.biblosSpeakBrailleInfo()

	def script_sayDocumentInfo(self,gesture):
		global bHan
		bHan.typedWord = []
		scriptCount=scriptHandler.getLastScriptRepeatCount()
		if scriptCount==0:
			bHan.biblosSpeakDocumentInfo(0)
		elif scriptCount==1:
			bHan.biblosSpeakDocumentInfo(1)


	__gestures = {
		"kb:f8": "sayChangeSelection",
		"kb:shift+upArrow": "sayChangeSelection",
		"kb:shift+downArrow": "sayChangeSelection",
		"kb:shift+leftArrow": "sayChangeSelection",
		"kb:shift+rightArrow": "sayChangeSelection",
		"kb:shift+home": "sayChangeSelection",
		"kb:shift+end": "sayChangeSelection",
		"kb:shift+pageUp": "sayChangeSelection",
		"kb:shift+pageDown": "sayChangeSelection",
		"kb:shift+control+leftArrow": "sayChangeSelection",
		"kb:shift+control+rightArrow": "sayChangeSelection",
		"kb:shift+control+upArrow": "sayChangeSelection",
		"kb:shift+control+downArrow": "sayChangeSelection",
		"kb:shift+home": "sayChangeSelection",
		"kb:shift+end": "sayChangeSelection",
		"kb:shift+control+home": "sayChangeSelection",
		"kb:shift+control+end": "sayChangeSelection",
		"kb:control+x": "cutToClipboard",
		"kb:control+c": "copyToClipboard",
		"kb:control+v": "pasteFromClipboard",
		"kb:f1": "sayF1Key",
		"kb:f2": "sayLine",
		"kb:shift+f2": "sayLine",
		"kb:f3": "sayLine",
		"kb:shift+f3": "sayLine",
		"kb:backspace": "sayBackspaceKey",
		"kb:control+pageUp": "sayLine",
		"kb:control+pageDown": "sayLine",
		"kb:enter": "sayEnterKey",
		"kb:shift+enter": "sayEnterKey",
		"kb:control+enter": "sayEnterKey",
		"kb:tab": "sayTabKey",
		"kb:shift+tab": "sayTabKey",
		"kb:NVDA+f": "sayTextInfo",
		"kb:NVDA+downArrow": "sayAll",
		"kb:alt+delete": "sayDocumentInfo",
		"kb:control+delete": "sayWord",
	}


# Developer believe that a comment can really explain the code. I'm going to cut my hair